/*
 *  linux/drivers/block/ide-tivo.c	Version 1.01  Nov  25, 1996
 *
 *  Copyright (C) 1994-1996  Linus Torvalds & authors (see below)
 */

/*
 *  Maintained by Mark Lord  <mlord@pobox.com>
 *            and Gadi Oxman <gadio@netvision.net.il>
 *
 * This is the IDE/ATA disk driver, as evolved from hd.c and ide.c.
 *
 *  From hd.c:
 *  |
 *  | It traverses the request-list, using interrupts to jump between functions.
 *  | As nearly all functions can be called within interrupts, we may not sleep.
 *  | Special care is recommended.  Have Fun!
 *  |
 *  | modified by Drew Eckhardt to check nr of hd's from the CMOS.
 *  |
 *  | Thanks to Branko Lankester, lankeste@fwi.uva.nl, who found a bug
 *  | in the early extended-partition checks and added DM partitions.
 *  |
 *  | Early work on error handling by Mika Liljeberg (liljeber@cs.Helsinki.FI).
 *  |
 *  | IRQ-unmask, drive-id, multiple-mode, support for ">16 heads",
 *  | and general streamlining by Mark Lord (mlord@pobox.com).
 *
 *  October, 1994 -- Complete line-by-line overhaul for linux 1.1.x, by:
 *
 *	Mark Lord	(mlord@pobox.com)		(IDE Perf.Pkg)
 *	Delman Lee	(delman@mipg.upenn.edu)		("Mr. atdisk2")
 *	Scott Snyder	(snyder@fnald0.fnal.gov)	(ATAPI IDE cd-rom)
 *
 *  This was a rewrite of just about everything from hd.c, though some original
 *  code is still sprinkled about.  Think of it as a major evolution, with
 *  inspiration from lots of linux users, esp.  hamish@zot.apana.org.au
 *
 * Version 1.00		move disk only code from ide.c to ide-tivo.c
 *			support optional byte-swapping of all data
 * Version 1.01		fix previous byte-swapping code
 */

#undef REALLY_SLOW_IO		/* most systems can safely undef this */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/errno.h>
#include <linux/hdreg.h>
#include <linux/genhd.h>
#include <linux/malloc.h>
#include <linux/random.h>

#include <asm/byteorder.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/mmu_context.h>
#include <asm/process.h>

#include "ide.h"
#include <linux/ide-tivo.h>
#include <linux/tivo-private.h>
#include <linux/tivo-diskmon.h>

#ifdef IDETIVO_PROFILE
#define PROFILE_ACCUM(mode) \
switch (mode) { \
 case ALL_ELEV: \
     hwif->tivo->elev++; \
     break; \
 case RT_ELEV: \
     hwif->tivo->elev_rt++; \
     break; \
 case RT_ONLY: \
     hwif->tivo->rt_only++; \
     break; \
}
#else
#define PROFILE_ACCUM(mode)
#endif 


/*
#define DEBUG
#define DEBUG_TIVO
#define DEBUG_TIVO_CALL
#define DEBUG_TIVO_STATUS
#define DEBUG_TIVO_RETURN
#define TIVO_STATUS3
*/

#include <linux/tivo.h>

#ifndef TRUE
#define TRUE (1)
#endif
#ifndef FALSE
#define FALSE (!TRUE)
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MIN3(a,b,c) MIN(a,MIN(b,c))
#define ABSDIFF(a,b) ((a)<(b)?((b)-(a)):((a)-(b)))
#endif


/* ////////////////////////////////////////////////////////////////////
// 
// incomplete types
// 
//////////////////////////////////////////////////////////////////// */

/* incomplete types */
struct idetivo_request; 

/* ////////////////////////////////////////////////////////////////////
// 
// prototypes
// 
//////////////////////////////////////////////////////////////////// */

extern pte_t *findPTE(struct mm_struct *, unsigned long);
extern void debug(char *, long);

/* prototypes for functions in this source file... */
static void idetivo_do_rw_disk (ide_drive_t *drive, struct idetivo_request
                                *rq);
static int idetivo_open (struct inode *inode, struct file *filp, ide_drive_t
                         *drive);
static void idetivo_release (struct inode *inode, struct file *filp,
                             ide_drive_t *drive);
static int idetivo_media_change (ide_drive_t *drive);
static void idetivo_pre_reset (ide_drive_t *drive);
static unsigned long idetivo_capacity (ide_drive_t  *drive);
static void idetivo_special (ide_drive_t *drive);
static int idetivo_readsectors(struct file *filp,
                               struct FsIovec *buf, int buf_len,
                               const struct FsIoRequest *request,
			       struct io_scramble *scramble); 
static int idetivo_writesectors(struct file *filp,
                                struct FsIovec *buf, int buf_len,
                                const struct FsIoRequest *request,
				struct io_scramble *scramble); 
static struct idetivo_request *idetivo_get_request(ide_drive_t *drive);
void idetivo_insert_request(ide_drive_t *drive, struct idetivo_request *req,
			    int athead);
void idetivo_service_request(ide_hwif_t *hwif);
static void idetivo_do_rw_cmd (ide_drive_t *drive, struct request *rq, unsigned long block);
static int idetivo_ioctl (ide_drive_t *drive, struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg);

void idetivo_update_stats (enum diskmon newState);
static void idetivo_update_longterm (void);
static void idetivo_normalize_stats (struct diskactivity *foo);
static int idetivo_install_watcher (struct file *filp, struct diskwatcher *watcher);
static void idetivo_remove_watcher (struct file *filp);

static long long now_usec(void);

/* Direct IO blocking hook */
extern void noaccess(unsigned long, int, int);
extern void pretlbFlush(void);

/* ////////////////////////////////////////////////////////////////////
// 
// data types
// 
//////////////////////////////////////////////////////////////////// */

static ide_module_t idetivo_module = {
    IDE_DRIVER_MODULE,
    idetivo_init,
    NULL
};

static struct lowlevel_operations idetivo_lowlevel_operations = {
    idetivo_readsectors, /* readsectors */
    idetivo_writesectors /* writesectors */
}; 

/*
 *	IDE subdriver functions, registered with ide.c
 */
static ide_driver_t idetivo_driver = {
    ide_tivo,		/* media */
    0,			/* busy */
    1,			/* supports_dma */
    0,			/* supports_dsc_overlap */
    NULL,			/* cleanup */
    idetivo_do_rw_cmd,		/* do_request */
    NULL,			/* end_request */
    idetivo_ioctl,		/* ioctl */
    idetivo_open,		/* open */
    idetivo_release,	/* release */
    idetivo_media_change,	/* media_change */
    idetivo_pre_reset,	/* pre_reset */
    idetivo_capacity,	/* capacity */
    idetivo_special		/* special */
};

/*
 * The following disk-time-usage stuff won't work correctly if there
 * is more than one disk controller.  Be warned.  Eventually this
 * data should be moved into the hwif so we can handle usage on
 * multiple controllers.  dplatt 12/13/2000
 */

static struct diskusage diskUsage = 
  { DISKUSAGE_VERSION };

static long long int diskTimeStart;
static int diskTimeState = DISKMON_IDLE;
static struct timer_list diskMonitorTimer;
static int diskMonitorTimerInstalled = 0;

static struct diskCondition {
   struct diskCondition *next;
   int triggered;
   int pid;
   struct file *filp;
   struct diskwatcher watch;
} *diskConditionHead = NULL;


/* ////////////////////////////////////////////////////////////////////
// 
// globals
// 
//////////////////////////////////////////////////////////////////// */

int (*idetivo_statistics_hook)(ide_drive_t *drive, 
			       struct idetivo_request *req,
			       int reset_statistics, 
			       struct request *rq);

extern void do_hwgroup_request (ide_hwgroup_t *hwgroup);

const int idetivo_version_supported = IDETIVO_DRIVER_VERSION;

/*
 * TiVo disk-usage-monitor routines
 */

void idetivo_stats_timer_expiry (unsigned long data)
{
   long int flags;
   struct diskCondition *cond;
   int i, trigger;
   save_flags(flags);
   cli();
   idetivo_update_stats(diskTimeState); /* burp it */
   idetivo_update_longterm();
   cond = diskConditionHead;
   while (cond) {
      if (!cond->triggered) {
	 trigger = 0;
	 for (i = 0 ; i < DISKMON_MAX_CLASSES && !trigger; i++) {
	    if ((cond->watch.thresholds.usage[i] > 0 &&
		 cond->watch.thresholds.usage[i] < diskUsage.shortterm.usage[i]) ||
		(cond->watch.thresholds.usage[i] < 0 &&
		 - cond->watch.thresholds.usage[i] > diskUsage.shortterm.usage[i])) {
	       trigger = 1;
	    }
	 }
	 if (trigger) {
	    cond->triggered = 1;
	    kill_proc(cond->pid, SIGIO, 1);
	 }
      }
      cond = cond->next;
   }
   diskMonitorTimer.expires  = jiffies + HZ / 2;
   add_timer(&diskMonitorTimer);
   restore_flags(flags);
}

void idetivo_update_stats (enum diskmon newState)
{
   long int flags;
   long long int timeNow, timeDelta;
   unsigned long addFactor, decayFactor, prod;
   int i;
   save_flags(flags);
   cli();
   if (!diskMonitorTimerInstalled) {
      diskMonitorTimer.function = idetivo_stats_timer_expiry;
      diskMonitorTimer.data     = 0;
      diskMonitorTimer.expires  = jiffies + HZ / 2;
      add_timer(&diskMonitorTimer);
      diskMonitorTimerInstalled = 1;
   }
   timeNow = now_usec();
   timeDelta = timeNow - diskTimeStart;
   diskTimeStart = timeNow;
   /*
    * Sanity check the time values, in case of timewarp
    * */
   if (timeDelta < 0 || timeDelta > 1000000) {
      timeDelta = 250000;
   }
   /*
    * Convert the time difference, in microseconds, down to a range
    * which fits nicely into 16 bits (it'll be in [0,62500))
    */
   addFactor = ((unsigned long) timeDelta) >> 4;
   decayFactor = 65535 - addFactor;
   for (i = 0 ; i < DISKMON_MAX_CLASSES; i++) {
      prod = (diskUsage.shortterm.usage[i] * decayFactor) >> 16;
      if (i == diskTimeState) {
	 prod += addFactor;
      }
      if (prod > 65535) {
	 prod = 65535; /* sanity clipping */
      }
      diskUsage.shortterm.usage[i] = prod;
   }
   diskTimeState = newState;
   restore_flags(flags);
   return;
}

#define DISKSCALEFACTOR 100

static void idetivo_update_longterm (void)
{
   int i;
   static int minutesThen = -1;
   int minutesNow;
   for (i = 0 ; i < DISKMON_MAX_CLASSES; i++) {
      diskUsage.mediumterm.usage[i] =
	((diskUsage.mediumterm.usage[i] * (DISKSCALEFACTOR - 1)) +
	 diskUsage.shortterm.usage[i]) / DISKSCALEFACTOR;
   }
   minutesNow = jiffies / (60 * HZ);
   if (minutesNow != minutesThen) {
      minutesThen = minutesNow;
      for (i = 0 ; i < DISKMON_MAX_CLASSES; i++) {
	 diskUsage.longterm.usage[i] =
	   ((diskUsage.longterm.usage[i] * (DISKSCALEFACTOR - 1)) +
	    diskUsage.mediumterm.usage[i]) / DISKSCALEFACTOR;
      }
   }
}

static void idetivo_normalize_stats (struct diskactivity *foo)
{
   int i;
   unsigned long sum = 0, sum2 = 0;
   for (i = 0 ; i < DISKMON_MAX_CLASSES; i++) {
      sum += foo->usage[i];
   }
   if (sum == 0) {
      foo->usage[DISKMON_IDLE] = 100;
      return;
   }
   for (i = DISKMON_IDLE+1; i < DISKMON_MAX_CLASSES ; i++) {
      foo->usage[i] = (foo->usage[i] * 100) / sum;
      sum2 += foo->usage[i];
   }
   foo->usage[DISKMON_IDLE] = 100 - sum2;
}

static int idetivo_install_watcher (struct file *filp,
				    struct diskwatcher *watcher)
{
   struct diskCondition *cond;
   long int flags;
   int found, ret, i;
   found = 0;
   save_flags(flags);
   cli();
   cond = diskConditionHead;
   while (cond && !found) {
      if (cond->filp == filp) {
	 found = 1;
      } else {
	 cond = cond->next;
      }
   }
   if (!found) {
      cond = (struct diskCondition *) kmalloc(sizeof *cond, GFP_KERNEL);
      if (!cond) {
	 restore_flags(flags);
	 return ENOMEM;
      }
      cond->filp = filp;
      cond->pid = current->pid;
      cond->next = diskConditionHead;
      diskConditionHead = cond;
   }
   if (copy_from_user((char *) &cond->watch, (char *) watcher, sizeof cond->watch)) {
      cond->triggered = 1;
      ret = EFAULT;
   } else {
      /* Rescale from percentage to 16-bit limits */
      for (i = 0 ; i < DISKMON_MAX_CLASSES ; i++) {
	cond->watch.thresholds.usage[i] *= 65535 / 100;
      }
      cond->triggered = 0;
      ret = 0;
   }
   restore_flags(flags);
   return ret;
}

static void idetivo_remove_watcher (struct file *filp)
{
   struct diskCondition *cond, *prev;
   long int flags;
   save_flags(flags);
   cli();
   cond = diskConditionHead;
   prev = NULL;
   while (cond && cond->filp != filp) {
      prev = cond;
      cond = cond->next;
   }
   if (cond) {
      if (prev) {
	 prev->next = cond->next;
      } else {
	 diskConditionHead = cond->next;
      }
      kfree(cond);
   }
}


/* ////////////////////////////////////////////////////////////////////
// 
// get_info_ptr
// 
//////////////////////////////////////////////////////////////////// */

/*
 * get_info_ptr() returns the (ide_drive_t *) for a given device number.
 * It returns NULL if the given device number does not match any present drives.
 */
static ide_drive_t *get_info_ptr (kdev_t i_rdev)
{
	int		major = MAJOR(i_rdev);
	unsigned int	h;

	for (h = 0; h < MAX_HWIFS; ++h) {
		ide_hwif_t  *hwif = &ide_hwifs[h];
		if (hwif->present && major == hwif->major) {
			unsigned unit = DEVICE_NR(i_rdev);
			if (unit < MAX_DRIVES) {
				ide_drive_t *drive = &hwif->drives[unit];
				if (drive->present)
					return drive;
			}
			break;
		}
	}
	return NULL;
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_init_driver_info
// 
//////////////////////////////////////////////////////////////////// */

struct idetivo_drive_info *idetivo_init_drive_info(void) {
    struct idetivo_drive_info *info;
    info = (struct idetivo_drive_info *) kmalloc(sizeof (struct idetivo_drive_info), GFP_KERNEL);
    if (!info) {
       return NULL;
    }
    info->state = STATE_UNKNOWN;
    info->family = FAMILY_UNKNOWN;

    info->back_to_back_writes = 0;
    info->back_to_back_limit = 2;   /* subject to later change... dplatt */

    memset(info->config, 0, sizeof(info->config));
    return(info);
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_init_driver_info
// 
//////////////////////////////////////////////////////////////////// */

struct idetivo_driver_info *idetivo_init_driver_info(void) {
    struct idetivo_driver_info *info;
    info = (struct idetivo_driver_info *) kmalloc(sizeof (struct idetivo_driver_info), GFP_KERNEL);
    if (!info) {
       return NULL;
    }
    info->version = IDETIVO_DRIVER_VERSION;
    info->request_head = NULL;
    info->request_head_rt = NULL;
    info->wait_for_request = NULL; 
    info->num_requests = 0;
    info->cur = NULL;
    info->state = STATE_UNKNOWN;
    info->head_pos.lba.block = 0;
    info->head_pos.chs.cylinder = 0;
    info->head_pos.chs.head = 0;
    info->head_pos.chs.sector = 0;

    memset(info->perf, 0, sizeof(info->perf));
    
#ifdef IDETIVO_PROFILE
    info->total_requests = 0;
    info->total_bytes = 0;
    info->total_jiffies = 0; 
#endif 
    memset(info->request_pool, 0,
           sizeof(info->request_pool));
    return(info);
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_insert_request
// 
//////////////////////////////////////////////////////////////////// */

#define IDETIVO_IN_ORDER(r1, r2) ((r1)->sector < (r2)->sector)

void idetivo_insert_request(ide_drive_t *drive, struct idetivo_request *req, int athead) {
    struct idetivo_request *head;
    unsigned long flags;
    int inserted; 
#if 0
   int counter;
#endif

    save_flags(flags);
    cli(); 
    TIVO_CALL("idetivo_insert_request"); 
    /* Insert into the elevator alg. ordered list */
    head = IDE_TIVO_INFO->request_head ;
    if (athead) {
       req->next = head;
       IDE_TIVO_INFO->request_head = req;
    } else if (!head) {
        IDE_TIVO_INFO->request_head = req;
        req->next = NULL;
    } else {
        while (head->next) { 
            if ((IDETIVO_IN_ORDER(head, req) ||
                 !IDETIVO_IN_ORDER(head, head->next)) &&
                (IDETIVO_IN_ORDER(req, head->next)))
                break;
            head = head->next;
        }
        req->next = head->next;
        head->next = req; 
    }

    inserted = 0; 
    /* Insert into the deadline ordered list. */
    head = IDE_TIVO_INFO->request_head_rt;
    if (!head ||
        (head && req->deadline && (!head->deadline ||
				   (head->deadline >= req->deadline)))) {
      IDE_TIVO_INFO->request_head_rt = req;
      if (head)
	req->next_rt = head;
      else
	req->next_rt = NULL;
    } else {
      while (head->next_rt && head->next_rt->deadline) {
	if (req->deadline && head->next_rt->deadline >= req->deadline)
	  break;
	head = head->next_rt;
      }
      req->next_rt = head->next_rt;
      head->next_rt = req;
    }
#if 0
    head = IDE_TIVO_INFO->request_head;
    counter = 0; 
    while(head && counter < REQUEST_POOL_SIZE) {
        printk("sc=(%ld,%ld) ", head->sector, (long)head->deadline);
        head = head->next;
        counter++; 
    }
    printk("\n"); 

    head = IDE_TIVO_INFO->request_head_rt;
    counter = 0; 
    while(head && counter < REQUEST_POOL_SIZE) {
        printk("rt=(%ld) ", (long)head->deadline);
        head = head->next_rt;
        counter++; 
    }
    printk("\n");
#endif 
    restore_flags(flags);
    }

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_get_request_wait
// 
//////////////////////////////////////////////////////////////////// */

static struct idetivo_request *idetivo_get_request_wait(ide_drive_t *drive) {
    struct wait_queue wait = { current, NULL };
    struct idetivo_request *req;

    req = idetivo_get_request(drive);
    if (req) return(req);

    printk(KERN_WARNING "no free requests...\n"); 
    add_wait_queue(&IDE_TIVO_INFO->wait_for_request, &wait); 
    for (;;) {
        req = idetivo_get_request(drive);
        if (req)
            break;
       /*
	* Looks to me as if we could fall into a race condition here,
	* if we manage to put ourself to sleep after somebody's awoken
	* us from the queue we put ourselves on up above.  We might
	* never wake up...   dplatt
	*/
        idetivo_service_request(drive->hwif);
        current->state = TASK_UNINTERRUPTIBLE;
        schedule();
    };
        
    remove_wait_queue(&IDE_TIVO_INFO->wait_for_request, &wait); 
    return(req); 
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_get_request
// 
//////////////////////////////////////////////////////////////////// */

/* Get an empty request from the request poool */ 
static struct idetivo_request *idetivo_get_request(ide_drive_t *drive) {
    int i;
    unsigned long flags;
    
    save_flags(flags);
    cli(); 
    for (i = 0; i < REQUEST_POOL_SIZE; i++) {
        if (IDE_TIVO_INFO->request_pool[i].status == IDETIVO_FREE) {
            struct idetivo_request *req;
            req = &IDE_TIVO_INFO->request_pool[i];
	    req->drive = drive;
            req->status = IDETIVO_BUSY;
            req->wait = NULL;
            req->next = NULL;
            req->next_rt = NULL; 
#ifdef TIVO_STATUS3
            printk(KERN_WARNING "using free request number %d (0x%lx)\n",
                   i, (long)&IDE_TIVO_INFO->request_pool[i]);
#endif /* TIVO_STATUS */
            restore_flags(flags);
            return(req);
        }
    }
    restore_flags(flags);
    return(NULL); 
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_finish_request
// 
//////////////////////////////////////////////////////////////////// */

static void idetivo_release_pages(FsIovec *req, int nreq) {
    int i;
    for (i = 0 ; i < nreq ; i++) {
        mem_map[MAP_NR(req[i].pb)].count--;
    }
}

static void idetivo_finish_request(ide_drive_t *drive,
                                   struct idetivo_request *which) {

    TIVO_CALL("idetivo_finish_request"); 

    // free up the memory that was allocated in copy_and_translate(). 
    idetivo_release_pages(which->state.head, which->state.head_len);
    kfree(which->state.head);
    which->status = IDETIVO_FREE;

    if (!IDE_TIVO_INFO) {
        printk(KERN_WARNING "NO TIVO INFO STRUCT!\n");
        return; 
    }

    if (!which->task) {
        printk(KERN_WARNING "no current task\n");
        return;
    }

    if (!drive) {
        printk(KERN_WARNING "no drive ptr\n");
        return;
    }
    
/*    IDE_TIVO_INFO->cur = NULL; */

    /* wake up anyone who's possibly waiting for a free request */
    wake_up(&IDE_TIVO_INFO->wait_for_request);

    wake_up_process(which->task); 

    if (drive->select.b.lba) {
        IDE_TIVO_INFO->head_pos.lba.block = which->sector + which->num_sectors; 
    } else {
        printk(KERN_WARNING "drive position CHS unknown\n"); 
        /* Get CHS position from drive */
    }
   
   /*
    * add some entropy to the pool
    */
   
   add_blkdev_randomness(HWIF(drive)->major);
   
} 
/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_remove_request
// 
//////////////////////////////////////////////////////////////////// */

static void idetivo_remove_request(struct idetivo_request *which) {
    struct idetivo_request *tmp;
    unsigned long flags;
    int deleted;
    ide_drive_t *drive = which->drive;
    
    save_flags(flags);
    cli();
/* remove from the elevator list */
    deleted = 0; 
    tmp = IDE_TIVO_INFO->request_head;
    if (tmp == which) {
        IDE_TIVO_INFO->request_head = tmp->next;
        deleted = 1;
    } else while (tmp) {
        if (tmp->next == which) {
            tmp->next = which->next;
            deleted = 1;
            which->next = NULL;
            break; 
        }
        tmp = tmp->next;
    }
    if (!deleted) {
        printk(KERN_WARNING "(%ld, %ld) not deleted from elev list\n",
               which->sector, (long)which->deadline);
    }

    /* remove from the realtime list */
    deleted = 0; 
    tmp = IDE_TIVO_INFO->request_head_rt;
    if (tmp == which) {
        IDE_TIVO_INFO->request_head_rt = tmp->next_rt;
        deleted = 1;
    } else while (tmp) {
        if (tmp->next_rt == which) { 
            tmp->next_rt = which->next_rt;
            deleted = 1;
            which->next_rt = NULL; 
            break;
        }
        tmp = tmp->next_rt;
    }
    if (!deleted) {
        printk(KERN_WARNING "(%ld, %ld) not deleted from rt list\n",
               which->sector, (long)which->deadline);
    }

    restore_flags(flags);
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_cancel_request
// 
//////////////////////////////////////////////////////////////////// */

void idetivo_cancel_request(ide_drive_t *drive,
                            struct idetivo_request *which) {
    IDE_TIVO_INFO->cur = NULL;
    which->status = IDETIVO_DONE;

    /* Wake up the calling process */
    wake_up(&which->wait);
    
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_release_request
// 
//////////////////////////////////////////////////////////////////// */

void idetivo_release_request(ide_drive_t *drive, 
                             struct idetivo_request *which) {
    
    TIVO_CALL("idetivo_release_request");
#if 0
    printk(KERN_WARNING "request is 0x%lx\n",
           (long)which);
#endif /* TIVO_STATUS */

    /* check to see if this was called because we're done with a chunk,
       or because we're done with the whole thing... */
    if (which->status == IDETIVO_BUSY &&
	which->sector_progress < which->num_sectors) {
        unsigned long bytes;
        FsIovec *ciov; 
        /* we've only done one chunk... */
#if 0
        printk(KERN_WARNING "release: starting new chunk: old at %ld new at %ld (old size %d)\n",
               which->sector,
               which->sector + which->chunk_num_sectors,
               which->chunk_num_sectors);
#endif
        bytes = which->chunk_num_sectors * SECTOR_BYTES; 
        which->sector += which->chunk_num_sectors;
        which->scramble.recordOffset ++;
        which->chunk_num_sectors = MIN(which->chunk_num_sectors, 
				       which->num_sectors -
                                       which->sector_progress);
        which->chunk_sector_progress = 0;

        ciov = which->state.top; 
#if 0
        printk(KERN_WARNING " old buffer 0x%p for %ld\n",
               ciov->pb, ciov->cb);
#endif
        while (bytes) {
            unsigned long chunk;
            chunk = MIN(ciov->cb, bytes);
            if (chunk == ciov->cb) {
                ciov++;
            } else {
                ciov->cb -= chunk;
                ciov->pb += chunk;
            }
	    which->state.top = ciov;
	    which->state.curr_iovec = ciov;
	    which->state.curr_iovec_bytes = 0; 
	    bytes -= chunk;
        }

#if 0
        printk(KERN_WARNING " new buffer 0x%p for %ld\n",
               which->state.top->pb, which->state.top->cb);
#endif
        idetivo_do_rw_disk(which->drive, which);
        return; 
    }
    idetivo_update_stats(DISKMON_IDLE);
    IDE_TIVO_INFO->cur = NULL;
    which->status = IDETIVO_DONE;

#ifdef TIVO_STATUS3
    printk(KERN_WARNING "Waking up i/o sleeping process. (which=%p, which->wait=%p\n",which, which->wait);
#endif /* TIVO_STATUS */
    /* wake up the calling process */
    wake_up(&which->wait);
#ifdef TIVO_STATUS3
    printk(KERN_WARNING "Waking up request pool sleepers. (wait_for_request=%p)\n",HWIF(drive)->tivo->wait_for_request);
#endif /* TIVO_STATUS */
    TIVO_RETURN("idetivo_release_request", "normal"); 
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_estimate_time
// 
//////////////////////////////////////////////////////////////////// */

static long long idetivo_estimate_time(ide_hwif_t *hwif,
				       long current_sector,
                                       struct idetivo_request *req) {
    long long result;

    // Performance numbers for a Quantum EX13.6 mounted on G411-1 grommets.
    // These are worst case seek + rotation and data transfer times, from
    // actual data.

    // 19ms worst case seek (no rotation) 
#define SEEK 19000 
    // 5400rpm = 90rps = 11111us/rev
#define ROTATION 11111
    // this is our worst case DTR -- 20ms for 256 sectors
#define TRANSFER (20000/256)
    
    result =
        SEEK +                          // initial seek time 
        ROTATION +                      // initial rotation time 
        (TRANSFER * req->num_sectors) + // transfer time 
        ((req->num_sectors>>8) * ROTATION); // extra rotations 
    return (result);
#ifdef TIVO_STATUS3 
    printk(KERN_WARNING "idetivo_estimate_time returns %ld\n",
           (long)result);
#endif 
}

/* ////////////////////////////////////////////////////////////////////
// 
// now_usec
// 
//////////////////////////////////////////////////////////////////// */

static long long now_usec(void) {
    long long msec;
    struct timeval tv;
    do_gettimeofday(&tv);
    msec = (long long)tv.tv_sec * 1000000 + tv.tv_usec;
    return(msec); 
}


/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_choose_request
// 
//////////////////////////////////////////////////////////////////// */
                                 
/* Choose the next best request to service from the list of outstanding
   requests. */
static struct idetivo_request *idetivo_choose_request(ide_hwif_t *hwif) {
    struct idetivo_request *result;
    struct idetivo_request *temp;     
    long long endtime;
    long long thistime; 
    int violation;
    int legacy;
    enum { ALL_ELEV = 0, RT_ELEV = 1, RT_ONLY = 2} mode; 

    TIVO_CALL("idetivo_choose_request");
    mode = ALL_ELEV;
    result = hwif->tivo->request_head;

    /* Maybe there are no more requests... */
    if (!result) {
/*        printk(KERN_WARNING "idetivo_choose_request: normal NULL!\n"); */
        return(NULL);
    }
    endtime = now_usec() + 
        idetivo_estimate_time(hwif,
			      hwif->tivo->head_pos.lba.block,
                              result);
    if (result->deadline && endtime >= result->deadline)
        goto rt_only; 
    temp = hwif->tivo->request_head;
    violation = FALSE; 
    legacy = FALSE;
    while (temp) {
        long long thistime;

	if (temp->drive && temp->drive->queue) {
	  legacy = TRUE;
	}

        /* Only consider requests that aren't "this one" and ones
           that have a deadline... */
        if (!temp->deadline || temp == result) {
            temp = temp->next;
            continue;
        }
        thistime = endtime +
            idetivo_estimate_time(hwif,
				  result->sector + result->num_sectors,
                                  temp);
        if (temp->deadline <= thistime) {
#ifdef TIVO_STATUS3
            printk(KERN_WARNING "request (%ld,%d,%ld) will violate request (%ld,%d,%ld)\n",
                   result->sector, result->num_sectors,
                   result->deadline,
                   temp->sector, temp->num_sectors, temp->deadline);
#endif
            violation = TRUE;
            break;
        }
        temp = temp->next;
    }
    if (!violation && !legacy) {
        mode = ALL_ELEV; 
        PROFILE_ACCUM(mode);
        return(result);
    }

    /* At this point, there was a violation in the standard elevator
       algorithm, or there were some legacy I/Os in a drive queue which
       cause us to not want to do non-deadlined requests.  Try elevator 
       of only the realtime requests... */
    result = hwif->tivo->request_head;
    while (result && !result->deadline)
        result = result->next;
    if (!result) {
        if (!legacy) {
	  printk(KERN_WARNING "Oi! This should never happen\n");
	}
        mode = ALL_ELEV; 
        PROFILE_ACCUM(mode);
        return(hwif->tivo->request_head); 
    }
    endtime = now_usec() +
        idetivo_estimate_time(hwif,
			      hwif->tivo->head_pos.lba.block,
                              result);
    if (result->deadline && endtime >= result->deadline)
        goto rt_only; 
    temp = hwif->tivo->request_head;
    violation = FALSE;
    while (temp) {
        if (!temp->deadline || temp == result) {
            temp = temp->next;
            continue;
        }
        thistime = endtime +
            idetivo_estimate_time(hwif,
				  result->sector + result->num_sectors,
                                  temp);
        if (temp->deadline <= thistime) {
            violation = TRUE;
            break;
        }
        temp = temp->next;
    }
    if (!violation) {
        mode = RT_ELEV; 
        PROFILE_ACCUM(mode);
        return(result);
    }
 rt_only: 
    mode = RT_ONLY; 
    PROFILE_ACCUM(mode);
    return(hwif->tivo->request_head_rt); 
} 

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_service_request
// 
//////////////////////////////////////////////////////////////////// */

/* pick the next request off the request list, and start servicing it */
void idetivo_service_request(ide_hwif_t *hwif) {
    /* If there are requests being serviced, then there's no way we
       can fire off another one... */
    struct idetivo_request *cur;
    unsigned long flags; 
    TIVO_CALL("idetivo_service_request");

    save_flags(flags);
    cli(); 
    if (hwif->tivo->cur) {
        TIVO_STATUS("idetivo_service_request", "request still pending...");
        return;
    }
    cur = idetivo_choose_request(hwif);
    hwif->tivo->cur = cur;
{ void chkIdeChBusyLock(void);
  extern int from_ide_intr;
/*
 * ??? SCS: There shouldn't be a lock on the DMA channel, this is a diag check
 * My assumption is that if there is no current request on the hwif, the DMA
 * channel can't be busy.  This check is only needed if we came from ide_intr.
 */
if (from_ide_intr) chkIdeChBusyLock();
}
    if (cur) {
#ifdef IDETIVO_PROFILE
        cur->num_interrupts = 0; 
        cur->iostart = jiffies;
#endif
       /*
	* The higher-level code in ide_do_request() has already filled
	* in the hwgroup->drive pointer.  However, idetivo_choose_request
	* may have chosen a request for a different drive on this HWIF,
	* based on the realtime values.  As a result, we must override
	* the earlier setting of the drive pointer.  dcp 3/6/99
	*/
        HWGROUP(cur->drive)->drive = cur->drive;
        idetivo_remove_request(cur);
        restore_flags(flags); 
        idetivo_do_rw_disk(cur->drive, cur);
    } else {
        TIVO_STATUS("idetivo_service_request", "idetivo_choose_request returned NULL\n");
    }
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_bswap_data
// 
//////////////////////////////////////////////////////////////////// */

static void idetivo_bswap_data (void *buffer, int wcount)
{
    u16 *p = buffer;

    TIVO_CALL("idetivo_bswap_data");
    while (wcount--) {
        *p++ = *p << 8 | *p >> 8;
        *p++ = *p << 8 | *p >> 8;
    }
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_input_data
// 
//////////////////////////////////////////////////////////////////// */

static int idetivo_input_data (ide_drive_t *drive,
                                       struct idetivo_request *req) 
{
    char *temp = IDE_TIVO_INFO->temp_page;
    int bufCounter;
    int count;

#if 0 
    TIVO_CALL("idetivo_input_data");
#endif 
    bufCounter = 0;

    count = SECTOR_BYTES;  /* This will never be called on multiread
                              accesses, we have a spearate interrupt routine
                              for that one... */

    while (count) {
        int s;
        ide_input_data(drive, temp, count/4);
        if (drive->bswap)
            idetivo_bswap_data(temp, count/4);

        s = FsIovecTransfer(&req->state, temp, SECTOR_BYTES);
        if (s != SECTOR_BYTES) {
            printk(KERN_WARNING "idetivo_input_data: couldn't transfer a sector worth of data! (got %d)\n", s); 
        }
        count -= s; 
    };
    TIVO_RETURN2("idetivo_input_data", "normal"); 
    return(SECTOR_BYTES); 
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_output_data
// 
//////////////////////////////////////////////////////////////////// */

static int idetivo_output_data (ide_drive_t *drive,
                                        struct idetivo_request *req)
{
    char *temp = IDE_TIVO_INFO->temp_page;
    int count;

    TIVO_CALL2("idetivo_output_data");

    count = SECTOR_BYTES;

    while (count) {
        int s;
        /* Take the data from user space, and put it into our temporary
           buffer... */
        s = FsIovecTransfer(&req->state, temp, SECTOR_BYTES); 
        if (s != SECTOR_BYTES) {
            printk(KERN_WARNING "idetivo_output_data: couldn't transfer a sector's worth of data (got %d)\n", s);
        }
        count -= s; 
    }
    /* now take our temporary buffer, and write it out to the drive... */
    ide_output_data(drive, temp, SECTOR_WORDS); 
    return(SECTOR_BYTES); 
}

static inline void idedisk_input_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
	ide_input_data(drive, buffer, wcount);
	if (drive->bswap)
		idetivo_bswap_data(buffer, wcount);
}

static inline void idedisk_output_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
	if (drive->bswap) {
		idetivo_bswap_data(buffer, wcount);
		ide_output_data(drive, buffer, wcount);
		idetivo_bswap_data(buffer, wcount);
	} else
		ide_output_data(drive, buffer, wcount);
}

#if 0

/* ////////////////////////////////////////////////////////////////////
// 
// lba_capacity_is_ok
// 
//////////////////////////////////////////////////////////////////// */

/*
 * lba_capacity_is_ok() performs a sanity check on the claimed "lba_capacity"
 * value for this drive (from its reported identification information).
 *
 * Returns:	1 if lba_capacity looks sensible
 *		0 otherwise
 */
static int lba_capacity_is_ok (struct hd_driveid *id)
{
    unsigned long lba_sects   = id->lba_capacity;
    unsigned long chs_sects   = id->cyls * id->heads * id->sectors;
    unsigned long _10_percent = chs_sects / 10;

    TIVO_CALL("lba_capacity_is_ok");
    /* perform a rough sanity check on lba_sects:  within 10% is "okay" */
    if ((lba_sects - chs_sects) < _10_percent)
        return 1;	/* lba_capacity is good */

    /* some drives have the word order reversed */
    lba_sects = (lba_sects << 16) | (lba_sects >> 16);
    if ((lba_sects - chs_sects) < _10_percent) {
        id->lba_capacity = lba_sects;	/* fix it */
        return 1;	/* lba_capacity is (now) good */
    }
    return 0;	/* lba_capacity value is bad */
}

#endif

/* ////////////////////////////////////////////////////////////////////
// 
// read_intr
// 
//////////////////////////////////////////////////////////////////// */

/*
 * read_intr() is the handler for disk read interrupts
 */
static void read_intr (ide_drive_t *drive)
{
    byte stat;
    struct idetivo_request *req; 
#if 0
    printk(KERN_WARNING "entering read_intr\n");
#endif

    if (!IDE_TIVO_INFO) {
        printk(KERN_WARNING "ide-tivo read_intr no info pointer!\n");
        return; 
    }

    req = IDE_TIVO_INFO->cur;
    if (!req) {
        printk(KERN_WARNING "read_intr: no current request! YIKES!"); 
        return; 
    }

#ifdef IDETIVO_PROFILE
    req->num_interrupts++;
#endif 

    if (!OK_STAT(stat=GET_STAT(),DATA_READY,BAD_R_STAT)) {
        idetivo_cancel_request(drive,req);
/*        ide_error(drive, "read_intr", stat); */
        return;
    }

    if (!req ||
        !drive ||
        req->drive != drive) {
        printk(KERN_WARNING "Wrong drive pointer.\n");
        return; 
    }
    
#ifdef IDETIVO_PROFILE
    req->num_interrupts++;
#endif 

    req->progress += idetivo_input_data(drive, req);
    req->sector_progress++;
    req->chunk_sector_progress++; 

    if (req->chunk_sector_progress >= req->chunk_num_sectors)
    {
#ifdef TIVO_STATUS3
        printk(KERN_WARNING "I wanted %d bytes, and I have %d\n",
               req->num_sectors * SECTOR_BYTES,
               req->progress); 
        printk(KERN_WARNING "ending request... \n");
#endif /* TIVO_STATUS */
        idetivo_release_request(drive, req);
        TIVO_RETURN("read_intr", "end of i/o"); 
        return;
    }
    TIVO_RETURN("read_intr", "more data"); 
    ide_set_handler(drive, &read_intr, WAIT_CMD); 
}

/*
 * read_intr_legacy() is the handler for disk read/multread interrupts
 */
static void read_intr_legacy (ide_drive_t *drive)
{
	byte stat;
	int i;
	unsigned int msect, nsect;
	struct request *rq;

	if (!OK_STAT(stat=GET_STAT(),DATA_READY,BAD_R_STAT)) {
		ide_error(drive, "read_intr", stat);
		return;
	}
	msect = drive->mult_count;
read_next:
	rq = HWGROUP(drive)->rq;
	if (msect) {
		if ((nsect = rq->current_nr_sectors) > msect)
			nsect = msect;
		msect -= nsect;
	} else
		nsect = 1;
	idedisk_input_data(drive, rq->buffer, nsect * SECTOR_WORDS);
#ifdef DEBUG
	printk("%s:  read: sectors(%ld-%ld), buffer=0x%08lx, remaining=%ld\n",
		drive->name, rq->sector, rq->sector+nsect-1,
		(unsigned long) rq->buffer+(nsect<<9), rq->nr_sectors-nsect);
#endif
	rq->sector += nsect;
	rq->buffer += nsect<<9;
	rq->errors = 0;
	i = (rq->nr_sectors -= nsect);
	if ((rq->current_nr_sectors -= nsect) <= 0) {
	        idetivo_update_stats(DISKMON_IDLE);
		ide_end_request(1, HWGROUP(drive));
                add_blkdev_randomness(HWIF(drive)->major);
	}
   
	if (i > 0) {
		if (msect)
			goto read_next;
		ide_set_handler (drive, &read_intr_legacy, WAIT_CMD);
	}
}

/* ////////////////////////////////////////////////////////////////////
// 
// write_intr
// 
//////////////////////////////////////////////////////////////////// */

static void write_done_intr (ide_drive_t *drive)
{
    byte stat;
    struct idetivo_request *req;

    TIVO_CALL("write_done_intr");
    req = IDE_TIVO_INFO->cur;
    if (!req ||
        !drive ||
        req->drive != drive) {
        printk(KERN_WARNING "Wrong drive pointer.\n");
        return; 
    }
#if 0
    printk(KERN_WARNING "calling write_done_intr\n");
#endif 
    if (!OK_STAT(stat=GET_STAT(),DRIVE_READY,drive->bad_wstat)) {
        idetivo_cancel_request(drive,req); 
/*        ide_error(drive, "write_done_intr", stat); */
    }


    idetivo_release_request(drive,req);
}

/*
 * write_intr() is the handler for disk write interrupts
 */
static void write_intr (ide_drive_t *drive)
{
    byte stat;
    struct idetivo_request *req;

    TIVO_CALL("write_intr");
    req = IDE_TIVO_INFO->cur;

    if (!req ||
        !drive ||
        req->drive != drive) {
        printk(KERN_WARNING "Wrong drive pointer (or 0x%p or 0x%p).\n",
               req, drive);
        return; 
    }

    if (!OK_STAT(stat=GET_STAT(),DRIVE_READY,drive->bad_wstat)) {
        idetivo_cancel_request(drive,req); 
/*        ide_error(drive, "write_intr", stat); */
        return;
    }

    if (!IDE_TIVO_INFO) {
        printk(KERN_WARNING "NO TIVO INFO POINTER!\n");
        return; 
    }
    
#ifdef IDETIVO_PROFILE
    req->num_interrupts++;
#endif 

    /* We have to check the progress before we write any bytes, because
           bytes are actually transferred before the first time this
           interrupt happens.  So, we check that here... */
    if (req->chunk_sector_progress >= req->chunk_num_sectors) {
        idetivo_release_request(drive,req);
        return; 
    }
    req->progress += idetivo_output_data(drive, req);
    req->sector_progress++;
    req->chunk_sector_progress++;

    if (req->chunk_sector_progress >= req->chunk_num_sectors) {
#ifdef TIVO_STATUS3
        printk(KERN_WARNING "I wanted to write %d bytes, and I have %d\n",
               req->num_sectors * SECTOR_BYTES,
               req->progress);
        printk(KERN_WARNING "ending write request\n");
#endif /* TIVO_STATUS */
        ide_set_handler(drive, &write_done_intr, WAIT_CMD); 
        return; 
    }

    ide_set_handler(drive, &write_intr, WAIT_CMD); 
}

/*
 * write_intr_legacy() is the handler for disk write interrupts, oldstyle
 */
static void write_intr_legacy (ide_drive_t *drive)
{
	byte stat;
	int i;
	ide_hwgroup_t *hwgroup = HWGROUP(drive);
	struct request *rq = hwgroup->rq;

	if (OK_STAT(stat=GET_STAT(),DRIVE_READY,drive->bad_wstat)) {
#ifdef DEBUG
		printk("%s: write: sector %ld, buffer=0x%08lx, remaining=%ld\n",
			drive->name, rq->sector, (unsigned long) rq->buffer,
			rq->nr_sectors-1);
#endif
		if ((rq->nr_sectors == 1) ^ ((stat & DRQ_STAT) != 0)) {
			rq->sector++;
			rq->buffer += SECTOR_BYTES;
			rq->errors = 0;
			i = --rq->nr_sectors;
			--rq->current_nr_sectors;
			if (rq->current_nr_sectors <= 0)
		                idetivo_update_stats(DISKMON_IDLE);
				ide_end_request(1, hwgroup);
			if (i > 0) {
				idedisk_output_data (drive, rq->buffer, SECTOR_WORDS);
				ide_set_handler (drive, &write_intr_legacy, WAIT_CMD);
			}
			return;
		}
	}
	ide_error(drive, "write_intr", stat);
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_end_request
// 
//////////////////////////////////////////////////////////////////// */

void idetivo_end_request(void) {
    TIVO_CALL("idetivo_end_request"); 
}


/* ////////////////////////////////////////////////////////////////////
// 
// set_multmode_intr
// 
//////////////////////////////////////////////////////////////////// */

/*
 * set_multmode_intr() is invoked on completion of a WIN_SETMULT cmd.
 */
static void set_multmode_intr (ide_drive_t *drive)
{
    byte stat = GET_STAT();

    TIVO_CALL("set_multmode_intr");
    if (OK_STAT(stat,READY_STAT,BAD_STAT)) {
        drive->mult_count = drive->mult_req;
    } else {
        drive->mult_req = drive->mult_count = 0;
        drive->special.b.recalibrate = 1;
        (void) ide_dump_status(drive, "set_multmode", stat);
    }
}

/* ////////////////////////////////////////////////////////////////////
// 
// set_geometry_intr
// 
//////////////////////////////////////////////////////////////////// */

/*
 * set_geometry_intr() is invoked on completion of a WIN_SPECIFY cmd.
 */
static void set_geometry_intr (ide_drive_t *drive)
{
    byte stat = GET_STAT();

    TIVO_CALL("set_geometry_intr");
    if (!OK_STAT(stat,READY_STAT,BAD_STAT))
        ide_error(drive, "set_geometry_intr", stat);
}

/* ////////////////////////////////////////////////////////////////////
// 
// recal_intr
// 
//////////////////////////////////////////////////////////////////// */

/*
 * recal_intr() is invoked on completion of a WIN_RESTORE (recalibrate) cmd.
 */
static void recal_intr (ide_drive_t *drive)
{
    byte stat = GET_STAT();

    TIVO_CALL("recal_intr");
    if (!OK_STAT(stat,READY_STAT,BAD_STAT))
        ide_error(drive, "recal_intr", stat);
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_do_rw_disk
// 
//////////////////////////////////////////////////////////////////// */

/*
 * idetivo_do_rw_disk() issues READ and WRITE commands to a disk,
 * using LBA if supported, or CHS otherwise, to address sectors.
 * It also takes care of issuing special DRIVE_CMDs.
 */
static void idetivo_do_rw_disk (ide_drive_t *drive,
                                struct idetivo_request *req)
{
    enum diskmon newState;
    TIVO_CALL("idetivo_do_rw_disk");
#if 0
    printk(KERN_WARNING "do_rw_disk: cmd %d, %d sectors at %ld\n",
           req->cmd, req->chunk_num_sectors, req->sector);
#endif
   
    if (idetivo_statistics_hook && 
	(*idetivo_statistics_hook)(drive, req, 0, NULL)) {
       return; /* temporary holdoff */
    }

    OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
    /* when you request 256 sectors, you send "0" to the drive */
    if (req->chunk_num_sectors < MAX_SECTORS)
        OUT_BYTE(req->chunk_num_sectors,IDE_NSECTOR_REG);
    else
        OUT_BYTE(0,IDE_NSECTOR_REG);
       
    if (drive->select.b.lba) {
        unsigned long temp;
        temp = req->sector; 
#ifdef DEBUG
        printk("%s: %sing: LBAsect=%ld, sectors=%d\n",
               drive->name, (req->cmd==READ)?"read":"writ",
               temp, req->chunk_num_sectors);
#endif
        OUT_BYTE(temp,IDE_SECTOR_REG);
        OUT_BYTE(temp>>=8,IDE_LCYL_REG);
        OUT_BYTE(temp>>=8,IDE_HCYL_REG);
        OUT_BYTE(((temp>>8)&0x0f)|(drive->select.all),IDE_SELECT_REG);
    } else {
        unsigned int sect,head,cyl,track;
        printk(KERN_WARNING "unexpected CHS mode!\n"); 
        track = req->sector / drive->sect;
        sect  = req->sector % drive->sect + 1;
        OUT_BYTE(sect,IDE_SECTOR_REG);
        head  = track % drive->head;
        cyl   = track / drive->head;
        OUT_BYTE(cyl,IDE_LCYL_REG);
        OUT_BYTE(cyl>>8,IDE_HCYL_REG);
        OUT_BYTE(head|(drive->select.all),IDE_SELECT_REG);
#ifdef DEBUG
        printk("%s: %sing: CHS=%d/%d/%d, sectors=%d\n",
               drive->name, (req->cmd==READ)?"read":"writ", cyl,
               head, sect, req->chunk_num_sectors);
#endif
    }
    if (req->partition_flags & 1) {
       newState = (req->deadline && req->deadline != INFINITE_DEADLINE) ? 
	 DISKMON_MEDIA_DEADLINE : DISKMON_MEDIA;
    } else {
       newState = DISKMON_MFS;
    }
    idetivo_update_stats(newState);
    if (req->cmd == READ) {
        TIVO_STATUS("idetivo_do_rw_disk" , "READ"); 
#if  defined(CONFIG_BLK_DEV_TRITON)||defined(CONFIG_TCD_IDE_DMA)
       if (drive->using_dma && !(HWIF(drive)->dmaproc)(ide_dma_read,drive)) {
            TIVO_RETURN("idetivo_do_rw_disk", "READ: DMA proc caled"); 
            return;
        }
#endif /* CONFIG_BLK_DEV_TRITON */
        ide_set_handler(drive, &read_intr, WAIT_CMD);
        if (drive->mult_count) {
            printk(KERN_ERR "%s: MULTWRITE NOT ALLOWED ON TIVO.\n",
                   drive->name);
	    idetivo_update_stats(DISKMON_IDLE);
            return;
        }
        OUT_BYTE(WIN_READ, IDE_COMMAND_REG);
        return;
    } else if (req->cmd == WRITE) {
        TIVO_STATUS("idetivo_do_rw_disk" , "WRITE"); 
#if defined(CONFIG_BLK_DEV_TRITON)||defined(CONFIG_TCD_IDE_DMA)
       if (drive->using_dma && !(HWIF(drive)->dmaproc)(ide_dma_write,drive)) {
            TIVO_RETURN("idetivo_do_rw_disk", "WRITE: DMA proc caled"); 
            return;
        }
#endif /* CONFIG_BLK_DEV_TRITON */
        if (drive->mult_count) {
            printk(KERN_ERR "%s: MULTWRITE NOT ALLOWED ON TIVO.\n",
                   drive->name);
	    idetivo_update_stats(DISKMON_IDLE);
            return; 
        }
        ide_set_handler(drive, &write_intr, WAIT_CMD);
        OUT_BYTE(WIN_WRITE, IDE_COMMAND_REG);
        if (ide_wait_stat(drive, DATA_READY, drive->bad_wstat, WAIT_DRQ)) {
            printk(KERN_ERR "%s: no DRQ after issuing %s\n", drive->name,
                   drive->mult_count ? "MULTWRITE" : "WRITE");
            TIVO_RETURN("idetivo_do_rw_disk", "WRITE: waiting for DRQ."); 
	    idetivo_update_stats(DISKMON_IDLE);
            return;
        }
        if (!drive->unmask)
            cli();
        req->progress += idetivo_output_data(drive, req);
        req->sector_progress++;
        req->chunk_sector_progress++; 
        return;
    } 
    printk(KERN_ERR "%s: bad command: %d\n", drive->name, req->cmd);
    idetivo_end_request();
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_do_rw_cmd - legacy rq-oriented I/O
// 
//////////////////////////////////////////////////////////////////// */

static void idetivo_do_rw_cmd (ide_drive_t *drive, struct request *rq, unsigned long block)
{
#ifdef CONFIG_BLK_DEV_PROMISE
	int use_promise_io = 0;
#endif /* CONFIG_BLK_DEV_PROMISE */

        if (idetivo_statistics_hook && 
           (*idetivo_statistics_hook)(drive, NULL, 0, rq)) {
	   /* temporary holdoff */
	   return;
        }

	OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
	OUT_BYTE(rq->nr_sectors,IDE_NSECTOR_REG);
#ifdef CONFIG_BLK_DEV_PROMISE
	if (IS_PROMISE_DRIVE) {
		if (hwif->is_promise2 || rq->cmd == READ) {
			use_promise_io = 1;
		}
	}
	if (drive->select.b.lba || use_promise_io) {
#else /* !CONFIG_BLK_DEV_PROMISE */
	if (drive->select.b.lba) {
#endif /* CONFIG_BLK_DEV_PROMISE */
#ifdef DEBUG
		printk("%s: %sing: LBAsect=%ld, sectors=%ld, buffer=0x%08lx\n",
			drive->name, (rq->cmd==READ)?"read":"writ",
			block, rq->nr_sectors, (unsigned long) rq->buffer);
#endif
		OUT_BYTE(block,IDE_SECTOR_REG);
		OUT_BYTE(block>>=8,IDE_LCYL_REG);
		OUT_BYTE(block>>=8,IDE_HCYL_REG);
		OUT_BYTE(((block>>8)&0x0f)|(drive->select.all),IDE_SELECT_REG);
	} else {
		unsigned int sect,head,cyl,track;
		track = block / drive->sect;
		sect  = block % drive->sect + 1;
		OUT_BYTE(sect,IDE_SECTOR_REG);
		head  = track % drive->head;
		cyl   = track / drive->head;
		OUT_BYTE(cyl,IDE_LCYL_REG);
		OUT_BYTE(cyl>>8,IDE_HCYL_REG);
		OUT_BYTE(head|(drive->select.all),IDE_SELECT_REG);
#ifdef DEBUG
		printk("%s: %sing: CHS=%d/%d/%d, sectors=%ld, buffer=0x%08lx\n",
			drive->name, (rq->cmd==READ)?"read":"writ", cyl,
			head, sect, rq->nr_sectors, (unsigned long) rq->buffer);
#endif
	}
#ifdef CONFIG_BLK_DEV_PROMISE
	if (use_promise_io) {
		do_promise_io (drive, rq);
		return;
	}
#endif /* CONFIG_BLK_DEV_PROMISE */
	idetivo_update_stats(DISKMON_LEGACY);
	if (rq->cmd == READ) {
#if defined(CONFIG_BLK_DEV_TRITON)||defined(CONFIG_TCD_IDE_DMA)
		if (drive->using_dma && !(HWIF(drive)->dmaproc(ide_dma_read, drive)))
			return;
#endif /* CONFIG_BLK_DEV_TRITON */
		ide_set_handler(drive, &read_intr_legacy, WAIT_CMD);
	        if (drive->mult_count) {
	            printk(KERN_ERR "%s: MULTWRITE NOT ALLOWED ON TIVO.\n",
	                   drive->name);
		    idetivo_update_stats(DISKMON_IDLE);
        	    return;
	        }
	        OUT_BYTE(WIN_READ, IDE_COMMAND_REG);
		return;
	}
	if (rq->cmd == WRITE) {
#if defined(CONFIG_BLK_DEV_TRITON)||defined(CONFIG_TCD_IDE_DMA)
		if (drive->using_dma && !(HWIF(drive)->dmaproc(ide_dma_write, drive)))
			return;
#endif /* CONFIG_BLK_DEV_TRITON */
	        if (drive->mult_count) {
	            printk(KERN_ERR "%s: MULTWRITE NOT ALLOWED ON TIVO.\n",
        	           drive->name);
		    idetivo_update_stats(DISKMON_IDLE);
	            return; 
	        }
		OUT_BYTE(WIN_WRITE, IDE_COMMAND_REG);
		if (ide_wait_stat(drive, DATA_READY, drive->bad_wstat, WAIT_DRQ)) {
			printk(KERN_ERR "%s: no DRQ after issuing %s\n", drive->name,
				drive->mult_count ? "MULTWRITE" : "WRITE");
		        idetivo_update_stats(DISKMON_IDLE);
			return;
		}
		if (!drive->unmask)
			cli();
		ide_set_handler (drive, &write_intr_legacy, WAIT_CMD);
		idedisk_output_data(drive, rq->buffer, SECTOR_WORDS);
		return;
	}
	printk(KERN_ERR "%s: bad command: %d\n", drive->name, rq->cmd);
	idetivo_update_stats(DISKMON_IDLE);
	ide_end_request(0, HWGROUP(drive));
}


/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_open
// 
//////////////////////////////////////////////////////////////////// */

static int idetivo_open (struct inode *inode, struct file *filp, ide_drive_t *drive)
{
    TIVO_CALL("idetivo_open");
    MOD_INC_USE_COUNT;
    if (drive->removable && drive->usage == 1) {
        byte door_lock[] = {WIN_DOORLOCK,0,0,0};
        struct request rq;
        check_disk_change(inode->i_rdev);
        ide_init_drive_cmd (&rq);
        rq.buffer = door_lock;
        /*
         * Ignore the return code from door_lock,
         * since the open() has already succeeded,
         * and the door_lock is irrelevant at this point.
         */
        (void) ide_do_drive_cmd(drive, &rq, ide_wait);
    }
    filp->f_llop = &idetivo_lowlevel_operations; 
    return 0;
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_release
// 
//////////////////////////////////////////////////////////////////// */

static void idetivo_release (struct inode *inode, struct file *filp, ide_drive_t *drive)
{
    TIVO_CALL("idetivo_release");

    if (!drive || !filp || !inode) {
        printk(KERN_WARNING "idetivo_release: drive=%p, filp=%p, inode=%p\n",
               drive, filp, inode);
        return; 
    }

    idetivo_remove_watcher(filp);

    if (drive->removable && !drive->usage) {
        byte door_unlock[] = {WIN_DOORUNLOCK,0,0,0};
        struct request rq;
        invalidate_buffers(inode->i_rdev);
        ide_init_drive_cmd (&rq);
        rq.buffer = door_unlock;
        (void) ide_do_drive_cmd(drive, &rq, ide_wait);
    }
    filp->f_llop = NULL; 
    MOD_DEC_USE_COUNT;
}
   
   
/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_media_change
// 
//////////////////////////////////////////////////////////////////// */

static int idetivo_media_change (ide_drive_t *drive)
{
    TIVO_CALL("idetivo_media_change");
    return drive->removable;	/* if removable, always assume it was changed */
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_capacity
// 
//////////////////////////////////////////////////////////////////// */

/*
 * current_capacity() returns the capacity (in sectors) of a drive
 * according to its current geometry/LBA settings.
 */
static unsigned long idetivo_capacity (ide_drive_t  *drive)
{
    struct hd_driveid *id = drive->id;
    unsigned long capacity = drive->cyl * drive->head * drive->sect;

    TIVO_CALL("idetivo_capacity");
#ifdef ORIGINAL 
    drive->select.b.lba = 0;
    /* Determine capacity, and use LBA if the drive properly supports it */
     if (id != NULL && (id->capability & 2) && lba_capacity_is_ok(id)) {
        if (id->lba_capacity >= capacity) {
            capacity = id->lba_capacity;
            drive->select.b.lba = 1;
        }
    }
#endif
    drive->select.b.lba = 0;
    /* Determine capacity, and use LBA if the drive properly supports it */
#if 0
    printk(KERN_WARNING "id=%p, id->capability=%d\n",
           id, id?id->capability:0);
#endif 
    if (id != NULL && (id->capability & 2)) {
        capacity = id->lba_capacity;
        drive->select.b.lba = 1;
#if 0 
        printk(KERN_WARNING "enabling LBA mode\n");
#endif 
    }
#if 0 
    printk(KERN_WARNING "lba mode is %d\n",
           drive->select.b.lba);
#endif 
    return (capacity - drive->sect0);
}
	   
/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_special
// 
//////////////////////////////////////////////////////////////////// */
	  	   
static void idetivo_special (ide_drive_t *drive)
{
   special_t *s = &drive->special;

    TIVO_CALL("idetivo_special");
#ifdef TIVO_STATUS3
    printk(KERN_WARNING "idetivo_special, all is 0x%08x\n",
           s->all);
    if (s->b.set_geometry) 
        TIVO_STATUS("idetivo_special", "should set geometry");
    if (s->b.recalibrate)
        TIVO_STATUS("idetivo_special", "should set status");
    if (s->b.set_multmode)
        TIVO_STATUS("idetivo_special", "should set multmode");
    if (s->b.set_tune)
        TIVO_STATUS("idetivo_special", "should set tune (unimplemented)");
    if (s->b.reserved)
        TIVO_STATUS("idetivo_special", "reserved bits set, thats bad");
#endif 
    if (s->b.set_geometry) {
        TIVO_STATUS("idetivo_special", "setting drive geometry");
        s->b.set_geometry = 0;
        OUT_BYTE(drive->sect,IDE_SECTOR_REG);
        OUT_BYTE(drive->cyl,IDE_LCYL_REG);
        OUT_BYTE(drive->cyl>>8,IDE_HCYL_REG);
        OUT_BYTE(((drive->head-1)|drive->select.all)&0xBF,IDE_SELECT_REG);
        if (!IS_PROMISE_DRIVE)
            ide_cmd(drive, WIN_SPECIFY, drive->sect, &set_geometry_intr);
    }
    if (s->b.recalibrate) {
        TIVO_STATUS("idetivo_special", "recalibrating drive");
        s->b.recalibrate = 0;
        if (!IS_PROMISE_DRIVE)
            ide_cmd(drive, WIN_RESTORE, drive->sect, &recal_intr);
    }
    if (s->b.set_multmode) {
        TIVO_STATUS("idetivo_special", "setting multmode");
        s->b.set_multmode = 0;
        if (drive->id && drive->mult_req > drive->id->max_multsect)
            drive->mult_req = drive->id->max_multsect;
        if (!IS_PROMISE_DRIVE)
            ide_cmd(drive, WIN_SETMULT, drive->mult_req, &set_multmode_intr);
    }
    if (s->all) {
        int special = s->all;
        s->all = 0;
        printk(KERN_ERR "%s: bad special flag: 0x%02x\n", drive->name, special);
    }
}
   
static int idetivo_ioctl (ide_drive_t *drive, struct inode *inode, struct file *filp,
			  unsigned int cmd, unsigned long arg)
{
   struct diskusage tDiskUsage;
   long int flags;
   switch (cmd) {

    case HDIO_SET_STREAMEDIA_MODE:
      if (!suser()) return -EACCES;
      if (arg) {
	 drive->part[MINOR(filp->f_inode->i_rdev) & PARTN_MASK].driver_flags |= 1;
      } else {
	 drive->part[MINOR(filp->f_inode->i_rdev) & PARTN_MASK].driver_flags &= ~1;
      }
      return 0;
    case HDIO_GET_DISKMON:
      save_flags(flags);
      cli();
      memcpy(&tDiskUsage, &diskUsage, sizeof tDiskUsage);
      restore_flags(flags);
      idetivo_normalize_stats(&tDiskUsage.shortterm);
      idetivo_normalize_stats(&tDiskUsage.mediumterm);
      idetivo_normalize_stats(&tDiskUsage.longterm);
      if (copy_to_user((char *)arg, (char *)&tDiskUsage, sizeof diskUsage))
	  return -EPERM;
      return 0;
    case HDIO_SET_WATCHER:
      if (arg) {
	 return idetivo_install_watcher (filp, (struct diskwatcher *) arg);
      } else {
	 idetivo_remove_watcher(filp);
	 return 0;
      }
   }
   return -EPERM;
}
   

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_pre_reset
// 
//////////////////////////////////////////////////////////////////// */

static void idetivo_pre_reset (ide_drive_t *drive)
{
    TIVO_CALL("idetivo_pre_reset");
    drive->special.all = 0;
    drive->special.b.set_geometry = 1;
    drive->special.b.recalibrate  = 1;
    if (OK_TO_RESET_CONTROLLER)
        drive->mult_count = 0;
    if (!drive->keep_settings)
        drive->mult_req = 0;
    if (drive->mult_req != drive->mult_count)
        drive->special.b.set_multmode = 1;
    drive->using_dma = 1;
    printk("%s: ide-tivo re-enabled DMA\n", drive->name);
}

/* ////////////////////////////////////////////////////////////////////
// 
// print_vma_table
// 
//////////////////////////////////////////////////////////////////// */

static void print_vma_table(struct vm_area_struct *head) {
    printk(KERN_WARNING "0x%lx - 0x%lx flags=%d offset=%ld\n",
           head->vm_start, head->vm_end, head->vm_flags, head->vm_offset);
#if 0
    printk(KERN_WARNING
           "code_start=%lx code_end=%lx data_start=%lx data_end=%lx start_brk=%lx brk=%lx\n",
           head->vm_mm->start_code, head->vm_mm->end_code,
           head->vm_mm->start_data, head->vm_mm->end_data,
           head->vm_mm->start_brk, head->vm_mm->brk);
#endif 
    if (head->vm_next)
        print_vma_table(head->vm_next); 
}

/* ////////////////////////////////////////////////////////////////////
// 
// copy_and_translate
// 
// note: the 'rw' parameter means which way the *DMA* is going. Thus,
//       'reading' actually means *writing* on main memory.
//
//////////////////////////////////////////////////////////////////// */

struct FsIovec *copy_and_translate(struct FsIovec *buf, int buf_len,
                                   int *new_buf_len, int rw) {
    // if you call this function, it will kmalloc its result value, and
    // you have to kfree it yourself!!! 
    int buf_count;
    struct FsIovec *curr_iovec;
    int bytecount; 
    int totbytes;
    unsigned long curr_ptr;
    unsigned char *temp;
    int max_count;
    struct FsIovec *new_iovec;
    struct FsIovec *curr_new_iovec; 

    /* This loop will compute the maximum number of entries that will
       be in the new iovec list */
    max_count = 0; 
    buf_count = 0; 
    curr_iovec = buf;
    while(buf_count < buf_len) {
        /* This is computed as:
           1 entry for the "first half" of the data
           1 entry for the "second half" of the data.
           1 entry for each "middle" page worth of data. */
        max_count += 2 + curr_iovec->cb / PAGE_SIZE;
        curr_iovec++;
        buf_count++; 
    }

    new_iovec = kmalloc(max_count * sizeof(struct FsIovec), 
			current->rt_priority >= 10 ? GFP_NFS : GFP_KERNEL);
    if (!new_iovec)
        return(NULL);
    curr_new_iovec = new_iovec; 
    
    /* Here, we're actually filling up the new iovec list... */
    *new_buf_len = 0; 
    buf_count = 0; 
    curr_iovec = buf;
    totbytes = 0;
    while (buf_count < buf_len) {
        curr_ptr = (unsigned long)curr_iovec->pb;
        bytecount = 0;
        while ((long)curr_iovec->cb > bytecount) {
            extern unsigned long make_page_present(unsigned long, int);
            unsigned long phys_addr;
            int chunk_size;

#ifdef RESCHED_DURING_COPY_AND_TRANSLATE	   
	    if (current->policy == SCHED_OTHER && need_resched)
                       schedule();
#endif

	    // Make the page present, and *quickly* ref-count it.
            phys_addr = make_page_present(curr_ptr, rw) + (curr_ptr % PAGE_SIZE);
            temp = (unsigned char *) phys_to_virt(phys_addr); 

            // Keep the page around till we're done. We know the count is at
            // least 1 before this (as this process has a handle on the page).
            mem_map[MAP_NR(temp)].count++;

            // Check to see if the current pointer is page-aligned...
            if (curr_ptr % PAGE_SIZE) {
                // ...it wasn't page aligned
                chunk_size = MIN(PAGE_SIZE - (curr_ptr % PAGE_SIZE),
                                 curr_iovec->cb - bytecount);
            } else {
                // ...it is page aligned
                chunk_size = MIN(PAGE_SIZE, curr_iovec->cb - bytecount); 
            }

            curr_new_iovec->pb = temp;
            curr_new_iovec->cb = chunk_size;

            (*new_buf_len)++;
            curr_new_iovec++; 
            bytecount += chunk_size; 
            curr_ptr += chunk_size;
        }
        curr_iovec++;
        buf_count++; 
	totbytes += bytecount;
    }
    if (totbytes % 512 != 0) {
       printk(KERN_WARNING "copy_and_translate: totbytes %d is not 0 mod 512!\n", totbytes);
       debug("copy_and_translate: bad length", 0);
    }
    return(new_iovec); 
}

/* ////////////////////////////////////////////////////////////////////
// 
// previous versions of readsectors/writesectors interface brokenly
// interpreted the request deadline as a sign-extended 32-bit number
// of microseconds (which wraps every ~4300 seconds).  This is an
// attempt to handle the case where user code is similarly broken,
// which would have mostly worked under previous version of interface.
// 
//////////////////////////////////////////////////////////////////// */
static long long fixup_deadline(long long deadline)
{
    long long now_time;
    long dl_high;

    dl_high = (long)(deadline >> 32);
    if (deadline == 0 || (dl_high != 0 && dl_high != 0xffffffff)) {
        return deadline;
    }
    now_time = now_usec();
    if ((unsigned long)deadline < 0x40000000 &&
        (unsigned long)now_time > 0xc0000000) {
        now_time += 0x100000000;
    } else if ((unsigned long)now_time < 0x40000000 &&
               (unsigned long)deadline > 0xc0000000) {
        now_time -= 0x100000000;
    }
    return (now_time & 0xffffffff00000000) | (unsigned long)deadline;
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_readsectors
// 
//////////////////////////////////////////////////////////////////// */
static int idetivo_readsectors(struct file *filp,
                               struct FsIovec *buf, int buf_len, 
                               const struct FsIoRequest *request,
			       struct io_scramble *scramble) {
    struct idetivo_request *cur;
    struct FsIovec *translated;
    int translated_len; 
    ide_drive_t *drive = get_info_ptr(filp->f_inode->i_rdev);
    unsigned long flags;
    unsigned int ret;
    int i;
    long int start_sect, nr_sects;
    unsigned long int sector; 
    struct io_scramble scrambleData;
#ifdef CONFIG_XMON
    unsigned int dbp1 = 0, dbp2 = 0;
#endif

    TIVO_CALL("idetivo_readsectors");

    ret = 0; 

    memset(&scrambleData, 0, sizeof scrambleData);

    if (scramble) {
      if (copy_from_user((char *)&scrambleData, (char *) scramble, sizeof scrambleData)) {
	 memset(&scrambleData, 0, sizeof scrambleData);
      } else {
	 if (scrambleData.magic != IO_SCRAMBLE_MAGIC ||
	     scrambleData.version != IO_SCRAMBLE_VERSION_TIVO1) {
	    memset(&scrambleData, 0, sizeof scrambleData);
	 }
      }
    }

    if (!drive) {
        printk(KERN_WARNING "drive is NULL\n"); 
        return(-1);
    }

    if (!drive->id) {
        printk(KERN_WARNING "drive->id is NULL\n");
        return(-1);
    }
    
    if (!request) {
        printk(KERN_WARNING "request is NULL\n");
        return(-1); 
    }
   
    start_sect = drive->part[MINOR(filp->f_inode->i_rdev) & PARTN_MASK].start_sect;
    nr_sects = drive->part[MINOR(filp->f_inode->i_rdev) & PARTN_MASK].nr_sects;
    /* Offset for the beginning of this partition */
    sector = request->sector + start_sect; 

    /* Check to see if its off the end of this partition */
    if (sector > (start_sect + nr_sects) ||
        (sector + request->num_sectors) > (start_sect + nr_sects)) {
        printk(KERN_WARNING "idetivo_readsectors: Access beyond end of drive\n");
        return(-1); 
    }
#if 0 
    if (request->num_sectors > MAX_SECTORS) {
        /* Can't send more than 256 to the drive! */
        return(-1); 
    }
#endif 
#ifdef CACHETEST
    for (i = 0 ; i < buf_len ; i++) {
      if ((((unsigned long) buf[i].pb) & 15) ||
	  (buf[i].cb & 15)) {
	printk(KERN_WARNING "idetivo_readsectors: transfer not an integral cache line!\n");
	break;
      }
    }
#endif
   
    cur = idetivo_get_request_wait(drive);
    if (cur) {
        struct wait_queue wait = { current , NULL };
	memcpy(&cur->scramble, &scrambleData, sizeof cur->scramble);
        cur->cmd = READ;
        cur->drive = drive;
        cur->partition_flags = drive->part[MINOR(filp->f_inode->i_rdev) & PARTN_MASK].driver_flags;
        cur->sector = sector;
        cur->num_sectors = request->num_sectors; 
        cur->chunk_num_sectors = MIN(MAX_SECTORS,request->num_sectors);
        if (scrambleData.magic == IO_SCRAMBLE_MAGIC &&
	    scrambleData.sectorsPerTransfer > 0 &&
	    scrambleData.sectorsPerTransfer < cur->chunk_num_sectors) {
	   cur->chunk_num_sectors = scrambleData.sectorsPerTransfer;
	}

        /* protect direct IO chunk */
        for (i = 0 ; i < buf_len ; i++) {
            noaccess((unsigned long)buf[i].pb, buf[i].cb, 1);
        }
#ifdef CONFIG_XMON
        /* turn on to check for VM accesses in this spot ... */
        pretlbFlush(); /* catch bad puppies ... */
#endif

        translated = copy_and_translate(buf, buf_len, &translated_len, 1);
        if (!translated) {

            /* couldn't translate; put things back */
            for (i = 0 ; i < buf_len ; i++) {
                noaccess((unsigned long)buf[i].pb, buf[i].cb, 0);
            }

            return(-1); 
	}

#ifdef CONFIG_XMON
        // Set up a trap in case kernel hits this memory.  This should never
        // happen, but if it does, we want to know about it (and fault into the
        // debugger)
        {
            extern unsigned int xSetDataBp(unsigned int addr, int size, int rw);

            pgd_t *pgd;
            pmd_t *pmd;
            pte_t *pte;
            int addrIter;

            for (addrIter = 0; addrIter < 2; addrIter++) {
                unsigned long badaddr = (unsigned long) buf[0].pb;
                unsigned long physAddr;
                int baddelta;

                // Select some random addresses to try to catch the kernel.
                switch (addrIter) {
                  case 0:
                    baddelta = 352;
                    break;

                  default:
                    baddelta = 2160;
                    break;
                }
                while (baddelta >= buf[0].cb) {
                    baddelta >>= 1;
                    baddelta &= ~0xf;
                }
                badaddr += baddelta;

                pgd = pgd_offset(current->mm, badaddr);
                if (pgd_none(*pgd)) {
                    debug("No page dir?", 0);
                }
                pmd = pmd_offset(pgd, badaddr);
                if (pmd_none(*pmd)) {
                    debug("No page middle dir?", 0);
                }
                pte = pte_offset(pmd, badaddr);
                if (pte_none(*pte)) {
                    debug("No pte?", 0);
                }
                if (!pte_present(*pte)) {
                    debug("pte not present?", 0);
                }

                physAddr = pte_page(*pte) + (badaddr & 0xfff);
                switch (addrIter) {
                  case 0:
                    dbp1 = xSetDataBp(physAddr, 3, 0);
                    break;
                  case 1:
                    dbp2 = xSetDataBp(physAddr, 3, 0);
                    break;
                }
            }
        }
#endif

        FsIovecInit(&cur->state,
                    translated, translated_len, IOVEC_READ);

        cur->progress = 0;
        cur->sector_progress = 0;
        cur->chunk_sector_progress = 0; 
        cur->deadline = fixup_deadline(request->deadline);
        cur->next = NULL;
        cur->next_rt = NULL;
        cur->task = current;

#if 0 
        printk(KERN_WARNING "readsectors : %d sectors from %d\n",
               cur->num_sectors, cur->sector);
#endif 
        
        add_wait_queue(&cur->wait, &wait);  
        idetivo_insert_request(drive, cur, 0);
	do_hwgroup_request(HWGROUP(drive));
#ifdef TIVO_STATUS3
        printk(KERN_WARNING "sleeping on i/o request...(cur=%p, cur->wait=%p)\n",cur,cur->wait);
#endif /* TIVO_STATUS */
        save_flags(flags);
        cli();
        while (cur->status != IDETIVO_DONE) {
            cur->task->state = TASK_UNINTERRUPTIBLE;
            schedule();
        } 
	restore_flags(flags);        
        remove_wait_queue(&cur->wait, &wait); 
#ifdef TIVO_STAUTS3
        printk(KERN_WARNING "waking up from i/o request! (cur=%p, cur->wait=%p)\n",cur,cur->wait);
#endif /* TIVO_STATUS */

#ifdef IDETIVO_PROFILE
        IDE_TIVO_INFO->num_requests++;
        IDE_TIVO_INFO->total_jiffies += jiffies - cur->iostart;
        IDE_TIVO_INFO->total_bytes += SECTOR_BYTES * cur->num_sectors; 
        if (!(IDE_TIVO_INFO->num_requests % 100)) {
#if 0
            printk(KERN_WARNING "Last %d requests; %ld bytes, Time: %ld jiffies (%d,%d,%d)\n",
                   IDE_TIVO_INFO->num_requests,
                   (long)IDE_TIVO_INFO->total_bytes,
                   (long)IDE_TIVO_INFO->total_jiffies,
                   IDE_TIVO_INFO->elev,
                   IDE_TIVO_INFO->elev_rt,
                   IDE_TIVO_INFO->rt_only);
#endif 
            IDE_TIVO_INFO->elev = 0; 
            IDE_TIVO_INFO->elev_rt = 0; 
            IDE_TIVO_INFO->rt_only = 0; 
            IDE_TIVO_INFO->total_bytes = 0;
            IDE_TIVO_INFO->total_jiffies = 0;
            IDE_TIVO_INFO->num_requests = 0; 
        }
#endif /* IDETIVO_PROFILE */

        ret = cur->progress; 
        idetivo_finish_request(drive, cur);

#ifdef CONFIG_XMON
        // Disable debug break points we set
        {
            extern void xRemoveDataBp(unsigned int addr);

            xRemoveDataBp(dbp1);
            xRemoveDataBp(dbp2);
        }
#endif

        /* release protected region */
        for (i = 0 ; i < buf_len ; i++) {
            noaccess((unsigned long)buf[i].pb, buf[i].cb, 0);
        }

    } else {
        printk(KERN_WARNING "couldn't get a free request\n"); 
    }
    return(ret);
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_writesectors
// 
//////////////////////////////////////////////////////////////////// */

static int idetivo_writesectors(struct file *filp,
                                struct FsIovec *buf, int buf_len,
                                const struct FsIoRequest *request,
				struct io_scramble *scramble) {
    ide_drive_t *drive;
    struct idetivo_request *cur;
    struct FsIovec *translated;
    int translated_len;
    unsigned long flags;
    struct wait_queue wait = { current, NULL };
    unsigned int ret;
    long int start_sect, nr_sects;
    unsigned long int sector; 
    struct io_scramble scrambleData;

    TIVO_CALL("idetivo_writesectors");

    memset(&scrambleData, 0, sizeof scrambleData);

    if (scramble) {
      if (copy_from_user((char *)&scrambleData, (char *) scramble, sizeof scrambleData)) {
/*	 printk("writesectors bad key, fault\n"); */
	memset(&scrambleData, 0, sizeof scrambleData);
      } else {
	 if (scrambleData.magic != IO_SCRAMBLE_MAGIC ||
	     scrambleData.version != IO_SCRAMBLE_VERSION_TIVO1) {
	    memset(&scrambleData, 0, sizeof scrambleData);
	 }
      }
    }

    drive = get_info_ptr(filp->f_inode->i_rdev);
    if (!drive) {
        TIVO_STATUS("idetivo_writesectors", "drive is NULL");
        return(-1);
    }
    
    /* Offset for the beginning of this partition */
    start_sect = drive->part[MINOR(filp->f_inode->i_rdev) & PARTN_MASK].start_sect;
    nr_sects = drive->part[MINOR(filp->f_inode->i_rdev) & PARTN_MASK].nr_sects;

    /* Offset for the beginning of this partition */
    sector = request->sector + start_sect; 

    /* Check to see if its off the end of this partition */
    if (sector > (start_sect + nr_sects) ||
        (sector + request->num_sectors) > (start_sect + nr_sects)) {
        printk(KERN_WARNING "idetivo_writesectors: Access beyond end of drive\n");
        return(-1); 
    }

    cur = idetivo_get_request_wait(drive);
    if (!cur) {
        printk(KERN_WARNING "couldn't get a free request\n");
        return(-1); 
    }
    memcpy(&cur->scramble, &scrambleData, sizeof cur->scramble);
    cur->cmd = WRITE;
    cur->drive = drive;
    cur->partition_flags = drive->part[MINOR(filp->f_inode->i_rdev) & PARTN_MASK].driver_flags;
    cur->sector = sector;
    cur->num_sectors = request->num_sectors;
    cur->chunk_num_sectors = MIN(MAX_SECTORS, request->num_sectors);
    if (scrambleData.magic == IO_SCRAMBLE_MAGIC &&
	scrambleData.sectorsPerTransfer > 0 &&
        scrambleData.sectorsPerTransfer < cur->chunk_num_sectors) {
       cur->chunk_num_sectors = scrambleData.sectorsPerTransfer;
    }

#if 0
    printk(KERN_WARNING "this write chunk is %d of %d\n",
           cur->chunk_num_sectors, cur->num_sectors); 
#endif 

    translated = copy_and_translate(buf, buf_len, &translated_len, 0);
    if (!translated) {
        return(-1); 
    }

    FsIovecInit(&cur->state,
                translated, translated_len, IOVEC_WRITE);
        
    cur->progress = 0;
    cur->sector_progress = 0;
    cur->chunk_sector_progress = 0; 
    cur->deadline = fixup_deadline(request->deadline);
    cur->next = NULL;
    cur->next_rt = NULL;
    cur->task = current;

#if 0
    printk(KERN_WARNING "writesectors : %d sectors from %d\n",
           cur->num_sectors, cur->sector);
#endif 
    
    add_wait_queue(&cur->wait, &wait); 
    idetivo_insert_request(drive, cur, 0);
    do_hwgroup_request(HWGROUP(drive));
#ifdef TIVO_STATUS3
    printk(KERN_WARNING "sleeping on i/o complete\n");
#endif
    save_flags(flags);
    cli();
    while (cur->status != IDETIVO_DONE) {
        cur->task->state = TASK_UNINTERRUPTIBLE;
        schedule();
    } 
    restore_flags(flags);        
    remove_wait_queue(&cur->wait, &wait); 
#ifdef IDETIVO_PROFILE
    IDE_TIVO_INFO->num_requests++;
    IDE_TIVO_INFO->total_jiffies += jiffies - cur->iostart;
    IDE_TIVO_INFO->total_bytes += SECTOR_BYTES * cur->num_sectors; 
    if (!(IDE_TIVO_INFO->num_requests % 100)) {
#if 0
        printk(KERN_WARNING "Last %d requests; %ld bytes, Time: %ld jiffies (%d,%d,%d)\n",
               IDE_TIVO_INFO->num_requests,
               (long)IDE_TIVO_INFO->total_bytes,
               (long)IDE_TIVO_INFO->total_jiffies,
               IDE_TIVO_INFO->elev,
               IDE_TIVO_INFO->elev_rt,
               IDE_TIVO_INFO->rt_only);
#endif 
        IDE_TIVO_INFO->elev = 0; 
        IDE_TIVO_INFO->elev_rt = 0; 
        IDE_TIVO_INFO->rt_only = 0; 
        IDE_TIVO_INFO->total_bytes = 0;
        IDE_TIVO_INFO->total_jiffies = 0;
        IDE_TIVO_INFO->num_requests = 0; 
    }
#endif /* IDETIVO_PROFILE */

    ret = cur->progress; 
    idetivo_finish_request(drive, cur); 

    return(ret);
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_cleanup
// 
//////////////////////////////////////////////////////////////////// */

static int idetivo_cleanup (ide_drive_t *drive)
{
    TIVO_CALL("idetivo_cleanup");
    return ide_unregister_subdriver(drive);
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_identify_device
// 
//////////////////////////////////////////////////////////////////// */

static int idetivo_identify_device (ide_drive_t *drive)
{
    struct hd_driveid *id = drive->id;

    TIVO_CALL("idetivo_identify_device");
    if (id == NULL)
        return 0;

    /* SunDisk drives: force one unit */
    if (id->model[0] == 'S' && id->model[1] == 'u' && (drive->select.all & (1<<4)))
        return 1;

    return 0;
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_setup
// 
//////////////////////////////////////////////////////////////////// */

static void idetivo_setup (ide_drive_t *drive)
{
    struct hd_driveid *id = drive->id;
    unsigned long capacity, check;

    TIVO_CALL("idetivo_setup");
    if (id == NULL)
        return;

#ifdef NOTDEF
    /* disable dma for now... */
    drive->using_dma = 0;
#endif
   
    /* check for removable disks (eg. SYQUEST), ignore 'WD' drives */
    if (id->config & (1<<7)) {	/* removable disk ? */
        if (id->model[0] != 'W' || id->model[1] != 'D')
            drive->removable = 1;
    }

    /* SunDisk drives: treat as non-removable */
    if (id->model[0] == 'S' && id->model[1] == 'u')
        drive->removable = 0;

    /* Extract geometry if we did not already have one for the drive */
    if (!drive->cyl || !drive->head || !drive->sect) {
        drive->cyl     = drive->bios_cyl  = id->cyls;
        drive->head    = drive->bios_head = id->heads;
        drive->sect    = drive->bios_sect = id->sectors;
    }
    /* Handle logical geometry translation by the drive */
    if ((id->field_valid & 1) && id->cur_cyls && id->cur_heads
        && (id->cur_heads <= 16) && id->cur_sectors)
    {
        /*
         * Extract the physical drive geometry for our use.
         * Note that we purposely do *not* update the bios info.
         * This way, programs that use it (like fdisk) will
         * still have the same logical view as the BIOS does,
         * which keeps the partition table from being screwed.
         *
         * An exception to this is the cylinder count,
         * which we reexamine later on to correct for 1024 limitations.
         */
        drive->cyl  = id->cur_cyls;
        drive->head = id->cur_heads;
        drive->sect = id->cur_sectors;

        /* check for word-swapped "capacity" field in id information */
        capacity = drive->cyl * drive->head * drive->sect;
        check = (id->cur_capacity0 << 16) | id->cur_capacity1;
        if (check == capacity) {	/* was it swapped? */
            /* yes, bring it into little-endian order: */
            id->cur_capacity0 = (capacity >>  0) & 0xffff;
            id->cur_capacity1 = (capacity >> 16) & 0xffff;
        }
    }
    /* Use physical geometry if what we have still makes no sense */
    if ((!drive->head || drive->head > 16) && id->heads && id->heads <= 16) {
        drive->cyl  = id->cyls;
        drive->head = id->heads;
        drive->sect = id->sectors;
    }
    /* Correct the number of cyls if the bios value is too small */
    if (drive->sect == drive->bios_sect && drive->head == drive->bios_head) {
        if (drive->cyl > drive->bios_cyl)
            drive->bios_cyl = drive->cyl;
    }

    /* Check advanced features */

    printk (KERN_INFO "idinfo 82=%04X 83=%04X 85=%04X 86=%04X 87=%04X\n",
	    id->smart, id->cmdsets_supported,
	    id->word85, id->cmdsets_enabled, id->word87);

    drive->av_cmdset = 0;
    drive->SMART = 0;

    if ((id->smart != 0x0000 && id->smart != 0xFFFF) ||
	(id->cmdsets_supported != 0x0000 && id->cmdsets_supported != 0xFFFF)) {
      if (id->smart & 0x1) {
	drive->SMART = 1;
      }
      if (id->cmdsets_supported & 0x80) {
	drive->av_cmdset = 1;
      }
    }

    (void) idetivo_capacity (drive); /* initialize LBA selection */

    printk (KERN_INFO "%s: %.40s, %ldMB w/%dkB Cache, multiple %d, %sCHS=%d/%d/%d%s%s%s\n",
            drive->name, id->model, idetivo_capacity(drive)/2048L,
            id->buf_size/2,
            id->max_multsect,
            drive->select.b.lba ? "LBA, " : "",
            drive->bios_cyl, drive->bios_head, drive->bios_sect,
            drive->using_dma ? ", DMA" : "",
            drive->av_cmdset ? ", A/V" : "",
	    drive->SMART ? ", SMART" : "");

    /* Force the drive to never use multmode */
    drive->mult_count = 0;
    if (id->max_multsect) {
#ifdef TIVO_STATUS3
        printk(KERN_WARNING "idetivo_setup: max_multsect=%d\n", id->max_multsect);
#endif 
        drive->mult_req = INITIAL_MULT_COUNT;
        if (drive->mult_req > id->max_multsect)
            drive->mult_req = id->max_multsect;
        if (drive->mult_req || ((id->multsect_valid & 1) && id->multsect)) {
            TIVO_STATUS("idetivo_setup", "using multmode");
            drive->special.b.set_multmode = 1;
            TIVO_STATUS("idetivo_setup", "calling special");
            DRIVER(drive)->special(drive);
            drive->special.all = 0; 
        } else {
#ifdef TIVO_STATUS3
            TIVO_STATUS("idetivo_setup", "not using multmode");
            printk(KERN_WARNING "idetivo_setup: drive->mult_req=%d, id->multsect_valid=%d id->multsect=%d\n",
                   drive->mult_req, id->multsect_valid,
                   id->multsect); 
#endif 
        }
#if 0
        /* TiVo: sending the special request to the disk... */
        if (drive->driver != NULL)
            DRIVER(drive)->special(drive);
        /* make sure we don't do this special request twice... */
        drive->special.all = 0;
#endif 
    } else {
        TIVO_STATUS("idetivo_setup","max_multsect was 0 now!"); 
    }
}

/* ////////////////////////////////////////////////////////////////////
// 
// idetivo_init
// 
//////////////////////////////////////////////////////////////////// */

int idetivo_init (void)
{
    ide_drive_t *drive;
    int failed = 0;

    TIVO_CALL("idetivo_init");
    MOD_INC_USE_COUNT;
    while ((drive = ide_scan_devices (ide_tivo, NULL, failed++)) != NULL) {
        if (idetivo_identify_device (drive))
            continue;
        TIVO_STATUS("idetivo_init", "about to register subdriver"); 
        if (ide_register_subdriver (drive, &idetivo_driver, IDE_SUBDRIVER_VERSION)) {
            printk (KERN_ERR "ide-tivo: %s: Failed to register the driver with ide.c\n", drive->name);
            continue;
        }
        idetivo_setup(drive);
        if ((!drive->head || drive->head > 16) && !drive->select.b.lba) {
            printk(KERN_ERR "%s: INVALID GEOMETRY: %d PHYSICAL HEADS?\n", drive->name, drive->head);
            (void) idetivo_cleanup(drive);
            continue;
        }
        failed--;
    }
    ide_register_module(&idetivo_module);
    MOD_DEC_USE_COUNT;
    return 0;
}

#ifdef MODULE

/* ////////////////////////////////////////////////////////////////////
// 
// init_module
// 
//////////////////////////////////////////////////////////////////// */

int init_module (void)
{
    TIVO_CALL("init_module");
    return idetivo_init();
}

/* ////////////////////////////////////////////////////////////////////
// 
// cleanup_module
// 
//////////////////////////////////////////////////////////////////// */

void cleanup_module (void)
{
    ide_drive_t *drive;
    int failed = 0;

    TIVO_CALL("cleanup_module");
    while ((drive = ide_scan_devices (ide_tivo, &idetivo_driver, failed)) != NULL)
        if (idetivo_cleanup (drive)) {
            printk (KERN_ERR "%s: cleanup_module() called while still busy\n", drive->name);
            failed++;
        }
    ide_unregister_module(&idetivo_module);
}
#endif /* MODULE */


