/*
 *  linux/fs/proc/inmap.c
 *
 *  Copyright (C) 2001 TiVo Inc.
 *
 *  proc page mapping handling functions
 */

#include <asm/uaccess.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/mm.h>

#include <asm/pgtable.h>

static int proc_readinmap(struct inode *, struct file *, void *, filldir_t);
static int proc_lookupinmap(struct inode *,const char *,int,struct inode **);
static long inmap_read(struct inode *, struct file *, char *, unsigned long);
static int inmap_open(struct inode *, struct file *);
static void inmap_release(struct inode *, struct file *);

static struct file_operations inode_ops = {
	NULL,			/* lseek - default */
	NULL,			/* read - bad */
	NULL,			/* write - bad */
	proc_readinmap,		/* readdir */
	NULL,			/* poll - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	NULL			/* can't fsync */
};

/*
 * proc directories can do almost nothing..
 */
struct inode_operations proc_inmap_inode_operations = {
	&inode_ops,		/* default base directory file-ops */
	NULL,			/* create */
	proc_lookupinmap,	/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL			/* permission */
};

static struct file_operations file_ops = {
	NULL,		/* plog_lseek */
	inmap_read,
	NULL,
	NULL,		/* plog_readdir */
	NULL,	        /* plog_poll */
	NULL,		/* plog_ioctl */
	NULL,		/* mmap */
	inmap_open,
	inmap_release,
	NULL		/* can't fsync */
};

struct inode_operations proc_inmap_operations = {
	&file_ops,		/* default base directory file-ops */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL			/* permission */
};

/*
 * Given a named directory entry in the map array, fashion an inode
 * that gives access to it. Relies on code in proc_get_inode() that
 * recognizes this type and does (mostly) nothing.
 */
static int proc_lookupinmap(struct inode * dir, const char * name, int len,
	struct inode ** result)
{
	unsigned int		ino;
	int			pid;
	struct super_block *	sb;
	struct task_struct *	p;
	unsigned long		vmstart;
	unsigned long		vmend;
	char *			endp;
	struct vm_area_struct *	map;
	int			i;

	*result = NULL;
	if (!dir)
		return -ENOENT;

	sb = dir->i_sb;
	ino = dir->i_ino;
	pid = ino >> 16;
	ino &= 0x0000ffff;

	/* "shouldn't happen" */
	if (!pid ||
	    ino != PROC_PID_INMAP ||
	    !S_ISDIR(dir->i_mode)) {
		iput(dir);
		return -ENOENT;
	}

	/* handle special entries (. and ..) */
	if (!len ||
	    (name[0] == '.' && (len == 1 ||
	    		       (name[1] == '.' && len == 2)))) {
		if (len < 2) {
			*result = dir;
			return 0;
		}
		if (!(*result = proc_get_inode(sb, (pid << 16) +
					PROC_PID_INO, &proc_pid))) {
			iput(dir);
			return -ENOENT;
		}
		iput(dir);
		return 0;
	}
	iput(dir);

	/* "shouldn't happen" */
	vmstart = simple_strtoul(name, &endp, 16);
	if (!vmstart)
		return -ENOENT;
	vmend = simple_strtoul(++endp, 0, 16);
	if (!vmend)
		return -ENOENT;

	/* get task pointer, task might also have exited ... */
	for (i = 0 ; i < NR_TASKS ; i++)
		if ((p = task[i]) && p->pid == pid)
			break;
	if (!pid || i >= NR_TASKS)
		return -ENOENT;

	/* index to the proper map, and return an inode for it */
	for (map = p->mm->mmap, i = 0;
	     map && i <= 256;
	     map = map->vm_next, i++) {
		if (map->vm_start == vmstart &&
		    map->vm_end == vmend) {
			ino = (pid << 16) + (PROC_PID_INMAP_DIR << 8) + i;
			if ((*result = proc_get_inode(sb, ino, NULL)))
				return 0;
		}
	}

	/* "shouldn't happen" */
	return -ENOENT;
}

/*
 * Fashion directory entries for the map array.
 */
#define NUMBUF 20

static int proc_readinmap(struct inode * inode, struct file * filp,
	void * dirent, filldir_t filldir)
{
	struct vm_area_struct *	map;
	char			buf[NUMBUF];
	int			task_nr;
	struct task_struct *	p;
	int			pid;
	unsigned int		ino;
	int			fd;

	/* "shouldn't happen" */
	if (!inode || !S_ISDIR(inode->i_mode))
		return -EBADF;

	/* "shouldn't happen" */
	ino = inode->i_ino;
	pid = ino >> 16;
	ino &= 0x0000ffff;
	if (ino != PROC_PID_INMAP)
		return 0;

	/* handle special entries, always first two, "." and ".." */
	for (fd = filp->f_pos; fd < 2; fd++, filp->f_pos++) {
		unsigned long nino = inode->i_ino;
		if (fd)
			nino = (ino & 0xffff0000) | PROC_PID_INO;
		if (filldir(dirent, "..", 2, fd, nino) < 0)
			return 0;
	}

	/* find the appropriate task, and it might have gone away */
	task_nr = 1;
	for (;;) {
		if ((p = task[task_nr]) && p->pid == pid)
			break;
		if (++task_nr >= NR_TASKS)
			return 0;
	}

	/* index to the proper map */
	for (map = p->mm->mmap, fd = 2;
	     map && fd < filp->f_pos;
	     map = map->vm_next, fd++);
	if (!map)
		return 0;

	/* index to the proper map, and fashion a directory entry */
	for (;
	     map;
	     map = map->vm_next, filp->f_pos++) {
		sprintf(buf, "%08lx-%08lx", map->vm_start, map->vm_end);
		ino = (pid << 16) | (PROC_PID_INMAP_DIR << 8) | (filp->f_pos - 2);
		if (filldir(dirent, buf, 17, filp->f_pos, ino) < 0)
			break;

		/* filldir() might have slept, so we must re-validate "p" */
		if (p != task[task_nr] || p->pid != pid)
			break;
	}

	/* all done */
	return 0;
}


/*
 * Heart of the work. Format out each page entry in turn. Entries are
 * a fixed number of characters per line, like this:
 * 	1 char		- "_" page not allocated, ignore page number
 * 			  "p" page present
 * 			  "s" page on swap
 *   	1 char		- "r" page readable
 *   			  "w" page writeable
 *	1 char		- "c" page clean
 *			  "d" page dirty
 *	1 char		- "t" page touched
 *			  "_" page untouched
 * 	8 characters	- page number in page map, or swap info
 * 	1 char		- ASCII space
 *	2 chars		- page reference count
 *	1 char		- ASCII newline
 *
 * All status characters will be "_" is page not present
 * or on swap.
 *
 * That's 16 characters per line, magic!
 */
#define PAGE_FORMAT	"%c%c%c%c%08lx %2d\n"
#define LEN_FORMAT	16
#define pos2pagenr(x)	(x >> 4)
#define pos2charnr(x)	(x & 0xf)

static long inmap_read(
	struct inode *		inode,
	struct file *		file,
	char *			buf,
	unsigned long		len
) {
	char			rbuf[LEN_FORMAT+1];
	int			error;
	int			mapnr;
	int			pagenr;
	int			npages;
	int			i;
	struct vm_area_struct *	map;
	unsigned long		ino;
	int			pid;
	struct task_struct *	p;
	unsigned long		addr;
	int			task_nr;
	int			nread = len;

	/* bad read call */
	if (!buf || len < 0)
		return -EINVAL;

	if (!len)
		return 0;

	error = verify_area(VERIFY_WRITE, buf, len);
	if (error)
		return error;

	/* "shouldn't happen" */
	ino = inode->i_ino;
	pid = ino >> 16;
	if (((ino >> 8) & 0xff) != PROC_PID_INMAP_DIR)
		return 0;

	/* find task pointer, verify process hasn't gone away */
	task_nr = 1;
	for (;;) {
		if ((p = task[task_nr]) && p->pid == pid)
			break;
		if (++task_nr >= NR_TASKS)
			return 0;
	}

	/* find map number this node is for 
	   note things could have been rearranged since the inode
	   was fashioned */
	mapnr = ino & 0xff;
	map = p->mm->mmap;
	for (i = 0; map && mapnr != i; i++, map = map->vm_next);
	if (!map)
		return 0;

	/* find where we were in the page sequence */
	pagenr = pos2pagenr(file->f_pos);
	npages = (map->vm_end - map->vm_start) / PAGE_SIZE;

	addr = map->vm_start + (pagenr * PAGE_SIZE);
	while (pagenr < npages) {
		pgd_t *			pgd;
		pmd_t *			pmd;
		pte_t *			pte;
		char			state = '_';
		char			rw = '_';
		char			cd = '_';
		char			touched = '_';
		unsigned long		frame = 0;
		int			charnr;

		/* see if the page exists at all */
		pgd = pgd_offset(map->vm_mm, addr);
		if (pgd_none(*pgd))
			goto nopage;
		pmd = pmd_offset(pgd, addr);
		if (pmd_none(*pmd))
			goto nopage;
		pte = pte_offset(pmd, addr);
		if (pte_none(*pte))
			goto nopage;

		/* must be on swap */
		frame = pte_page(*pte);
		if (!pte_present(*pte)) {
			state = 's';
			goto nopage;
		}
		else {
			state = 'p';
		}

		/* check state bits */
		if (pte_write(*pte))
			rw = 'w';
		else
			rw = 'r';
		if (pte_dirty(*pte))
			cd = 'd';
		else
			cd = 'c';
		if (pte_young(*pte))
			touched = 't';

	nopage:
		/* format entry, start from last position */
		if (state == 'p')
			sprintf(rbuf, PAGE_FORMAT,
				state, rw, cd, touched, frame,
				mem_map[MAP_NR(frame)].count);
		else
			sprintf(rbuf, PAGE_FORMAT,
				state, rw, cd, touched, frame,
				0);
		charnr = pos2charnr(file->f_pos);
		for (;
		     charnr < LEN_FORMAT && len > 0;
		     charnr++, len--, file->f_pos++)
			put_user(rbuf[charnr], buf++);
		pagenr++;
		addr += PAGE_SIZE;
	}

	return nread - len;
}

static int inmap_open(struct inode * inode, struct file * file)
{
    return 0;
}

static void inmap_release(struct inode * inode, struct file * file)
{
}
